import importlib
import os
import sys
from functools import update_wrapper
import click


FOLDERS = (
    "configuration",
    "input",
    "log",
    "tmp",
    "output",
)


@click.group()
@click.option(
    "--id",
    default=None,
    show_default=True,
    help="Id of the analysis being undertaken. None (for local use) will cause a unique ID to be generated.",
)
@click.option(
    "--skip-checks/--no-skip-checks",
    default=False,
    is_flag=True,
    show_default=True,
    help="Skips the input checking. This can be a timesaver if you already checked "
    "data directories (especially if manifests are large).",
)
@click.option(
    "--log-level",
    default="info",
    type=click.Choice(["debug", "info", "warning", "error"], case_sensitive=False),
    show_default=True,
    help="Log level used for the analysis.",
)
@click.option(
    "--force-reset/--no-force-reset",
    default=True,
    is_flag=True,
    show_default=True,
    help="Forces a reset of analysis cache and outputs [For future use, currently not implemented]",
)
@click.option(
    "--configuration-values",
    type=click.Path(),
    default="<data-dir>/input/input_values.json",
    show_default=True,
    help="Source for configuration_values strand data.",
)
@click.option(
    "--configuration-manifest",
    type=click.Path(),
    default="<data-dir>/input/input_values.json",
    show_default=True,
    help="Source for configuration_manifest strand data.",
)
@click.option(
    "--input-values",
    type=click.Path(),
    default="<data-dir>/input/input_values.json",
    show_default=True,
    help="Source for input_values strand data.",
)
@click.option(
    "--input-manifest",
    type=click.Path(),
    default="<data-dir>/input/input_manifest.json",
    show_default=True,
    help="Source for input_manifest strand data.",
)
@click.option(
    "--output-dir",
    type=click.Path(),
    default="output",
    show_default=False,
    help="Directory to write outputs as files.",
)
@click.option(
    "--log-dir", type=click.Path(), default="logs", show_default=True, help="Path to the location of log files",
)
@click.pass_context
def octue_cli(
    ctx,
    id,
    skip_checks,
    log_level,
    force_reset,
    configuration_values,
    configuration_manifest,
    input_values,
    input_manifest,
    data_dir,
    input_dir,
    tmp_dir,
    output_dir,
    log_dir,
):
    """ Octue CLI, enabling a data service / digital twin to be run like a command line application.

    Provide sources of configuration and/or input data and run the app. A source can be:

    - A path (relative or absolute) to a directory containing a <strand>.json file (eg `path/to/dir`).
    - A path to a <strand>.json file (eg `path/to/configuration_values.json`).
    - A literal JSON string (eg `{"n_iterations": 10}`.

    """

    # We want to show meaningful defaults in the CLI help but unfortunately have to strip out the displayed values here
    if input_values.startswith("<data-dir>/"):
        input_dir = None  # noqa
    if log_dir.startswith("<data-dir>/"):
        log_dir = None  # noqa
    if output_dir.startswith("<data-dir>/"):
        output_dir = None  # noqa

    ctx.ensure_object(dict)
    ctx.obj["analysis"] = "VIBRATION"


def pass_analysis(f):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        return ctx.invoke(f, ctx.obj["analysis"], *args, **kwargs)

    return update_wrapper(new_func, f)


def octue_run(f):
    """ Decorator for the main `run` function which adds a command to the CLI and prepares analysis ready for the run
    """

    @octue_cli.command()
    @pass_analysis
    def run(*args, **kwargs):
        return f(*args, **kwargs)

    return update_wrapper(run, f)


def octue_version(f):
    """ Decorator for the main `version` function which adds a command to the CLI
    """

    @octue_cli.command()
    def version(*args, **kwargs):
        return f(*args, **kwargs)

    return update_wrapper(version, f)


def unwrap(fcn):
    """ Recurse through wrapping to get the raw function without decorators.
    """
    if hasattr(fcn, "__wrapped__"):
        return unwrap(fcn.__wrapped__)
    return fcn


class AppFrom:
    """ Context manager that allows us to temporarily add an app's location to the system path and
    extract its run function

    with AppFrom('/path/to/dir') as app:
        Runner().run(app)

    """

    def __init__(self, app_path="."):
        self.app_path = os.path.abspath(os.path.normpath(app_path))
        self.app_module = None

    def __enter__(self):
        sys.path.insert(0, self.app_path)
        self.app_module = importlib.import_module("app")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.app_path in sys.path:
            sys.path.remove(self.app_path)

    @property
    def run(self):
        """ Returns the unwrapped run function from app.py in the application's root directory
        """
        return unwrap(self.app_module.run)



import logging


class Loggable:
    """ Mixin to allow instantiation of a class with a logger, or by default use the module logger from that class

    ```
    class MyResource(Logged):
        def do_something(self):
            self.logger.info('write to a logger')

    MyResource().do_something()  # Log statements go to the default logger for the module in which MyResource is a member
    MyResource(logger=logging.getLogger("specific"))  # Log statements go to the specific logger.
    ```
    """

    def __init__(self, *args, logger=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.__init_logger__(logger)

    def __init_logger__(self, logger=None):
        """ Overload this in a subclass to initialise your own logger using class attributes
        """
        self._logger = logger or logging.getLogger(self.__class__.__module__)

    @property
    def logger(self):
        return self._logger



import logging
import os

from octue.exceptions import BrokenSequenceException, InvalidInputException, UnexpectedNumberOfResultsException
from octue.mixins import Identifiable, Loggable, Serialisable, Taggable
from octue.resources.datafile import Datafile
from octue.utils import isfolder


module_logger = logging.getLogger(__name__)


class Dataset(Taggable, Serialisable, Loggable, Identifiable):
    """ A representation of a dataset, containing files, tags, etc

    This is used to read a list of files (and their associated properties) into octue analysis, or to compile a
    list of output files (results) and their properties that will be sent back to the octue system.
    """

    def __init__(self, id=None, logger=None, tags=None, **kwargs):
        """ Construct a Dataset
        """
        super().__init__(id=id, logger=logger, tags=tags)
        self.files = kwargs.pop("files", list())
        self.__dict__.update(**kwargs)

        # TODO A much better way than relying on the current directory!
        self._path = os.path.abspath(f"./dataset-{self.id}")

    def append(self, *args, **kwargs):
        """ Add a data/results file to the manifest

        Usage:
            my_file = octue.DataFile(...)
            my_manifest.append(my_file)

            # or more simply
            my_manifest.append(**{...}) which implicitly creates the datafile from the starred list of input arguments

        TODO allow for appending a list of datafiles
        """
        if len(args) > 1:
            # Recurse to allow addition of many files at once
            for arg in args:
                self.append(arg, **kwargs)
        elif len(args) > 0:
            if not isinstance(args[0], Datafile):
                raise InvalidInputException(
                    'Object "{}" must be of class Datafile to append it to a Dataset'.format(args[0])
                )
            self.files.append(args[0])

        else:
            # Append a single file, constructed by passing the arguments through to DataFile()
            self.files.append(Datafile(**kwargs))

    def get_files(self, field_lookup, files=None, filter_value=None):
        """ Get a list of data files in a manifest whose name contains the input string

        Django field lookups: https://docs.djangoproject.com/en/3.1/ref/models/querysets/

        :parameter field_lookup: filter specifier. Inspired by django queryset filters which in turn mimic SQL syntax.
        Comprised of "<field_name>__<filter_kind>", so 'name__icontains' would perform case-insensitive matching on
        Datafile.name. See https://github.com/octue/octue-sdk-python/issues/7 Currently supported combinations:

            "name__icontains"
            "name__contains"
            "name__endswith"
            "tag__exact"
            "tag__startswith"
            "tag__endswith"
            "sequence__gt"

        :parameter files: List of files to filter. This allows multiple separate filters to be applied. By default the
        entire list of files in the dataset is used.
        :type files: list

        :parameter filter_value: Value, typically of the same type as the field, used to filter against.
        : type filter_value: number, str

        :return: results list of matching datafiles
        :rtype: list(Datafile)
        """

        # Search through the input list of files or by default all files in the dataset
        files = files or self.files

        # Frequent error of typing only a single underscore causes no results to be returned... catch it
        if "__" not in field_lookup:
            raise InvalidInputException("Field lookups should be in the form '<field_name>__'<filter_kind>")

        results = []
        for file in files:
            if field_lookup == "name__icontains" and filter_value.lower() in file.name.lower():
                results.append(file)
            if field_lookup == "name__contains" and filter_value in file.name:
                results.append(file)
            if field_lookup == "name__endswith" and file.name.endswith(filter_value):
                results.append(file)
            if field_lookup == "name__startswith" and file.name.startswith(filter_value):
                results.append(file)
            if field_lookup == "tag__exact" and filter_value in file.tags:
                results.append(file)
            if field_lookup == "tag__startswith" and file.tags.startswith(filter_value):
                results.append(file)
            if field_lookup == "tag__endswith" and file.tags.endswith(filter_value):
                results.append(file)
            if field_lookup == "tag__contains" and file.tags.contains(filter_value):
                results.append(file)
            if field_lookup == "sequence__notnone" and file.sequence is not None:
                results.append(file)

        return results

    def get_file_sequence(self, field_lookup, files=None, filter_value=None, strict=True):
        """ Get an ordered sequence of files matching a criterion

        Accepts the same search arguments as `get_files`.

        :parameter strict: If True, applies a check that the resulting file sequence begins at 0 and ascends uniformly
        by 1
        :type strict: bool

        :returns: Sorted list of Datafiles
        :rtype: list(Datafile)
        """

        results = self.get_files(field_lookup, files=files, filter_value=filter_value)
        results = self.get_files("sequence__notnone", files=results)

        def get_sequence_number(file):
            return file.sequence

        # Sort the results on ascending sequence number
        results.sort(key=get_sequence_number)

        # Check sequence is unique and sequential
        if strict:
            index = -1
            for result in results:
                index += 1
                if result.sequence != index:
                    raise BrokenSequenceException("Filtered file sequence numbers do not monotonically increase from 0")

        return results

    def get_file_by_tag(self, tag_string):
        """ Gets a data file from a manifest by searching for files with the provided tag(s)

        Gets exclusively one file; if no file or more than one file is found this results in an error.

        :param tag_string: if this string appears as an exact match in the tags
        :return: DataFile object
        """
        results = self.get_files("tag__exact", filter_value=tag_string)
        if len(results) > 1:
            raise UnexpectedNumberOfResultsException("More than one result found when searching for a file by tag")
        elif len(results) == 0:
            raise UnexpectedNumberOfResultsException("No files found with this tag")

        return results[0]

    @property
    def path(self):
        # Lazily make the folder if absent
        isfolder(self._path, make_if_absent=True)
        return self._path



import json

from octue.resources import Datafile
from octue.utils.encoders import OctueJSONEncoder
from .mandelbrot import mandelbrot


def fractal(analysis):
    """ Compute the heightmap of a fractal and output a data file containing visualisation data
    """

    # Call the 'mandel' function to compute the fractal. Here, we just treat 'mandel' as a legacy function, passing in
    # what we need from the analysis object and getting results back
    x, y, z = mandelbrot(
        analysis.configuration_values["width"],
        analysis.configuration_values["height"],
        analysis.configuration_values["x_range"],
        analysis.configuration_values["y_range"],
        analysis.configuration_values["n_iterations"],
    )

    # The best thing to do with a fractal is to plot it. Let's do that.
    #
    # Figures are based on the Plotly figure spec, and you can use any of the plotly python tools here to create, show
    # and develop your figures while running locally.
    #
    # The beauty of the plotly system is that it's entirely json based, which means we can deliver figures and
    # dashboards as outputs (or monitors updated in real time) from digital twins and data services.
    #
    # Often, it's quicker to create the data and layout yourself than to use plotly's graph_objects library, so we do
    # that here:
    data = {
        "x": x,
        "y": y,
        "z": z,
        "colorscale": analysis.configuration_values["color_scale"],
        "type": "surface",
    }
    layout = {
        "title": f"Mandelbrot set with {analysis.configuration_values['n_iterations']} iterations",
        "width": analysis.configuration_values["width"],
        "height": analysis.configuration_values["height"],
    }

    # We'll add some tags, which will help to improve searchability and allow
    # other apps, reports, users and analyses to automatically find figures and
    # use them.
    #
    # Get descriptive with tags... they are whitespace-delimited and colons can be
    # used to provide subtags. Tags are case insensitive, and accept a-z, 0-9,
    # hyphens and underscores (which can be used literally in search and are also
    # used to separate words in natural language search). Other special characters
    # will be stripped.
    tags = "contents:fractal:mandelbrot type:figure:surface"

    # Get the output dataset which will be used for storing the figure file(s)
    output_dataset = analysis.output_manifest.get_dataset("fractal_figure_files")

    # Create a Datafile to hold the figure. We could put it in the current directory (by leaving local_path_prefix
    # unspecified) but it makes sense to put it in a folder specific to this output dataset - doing so avoids any race
    # conditions arising (if other instances of this application are running at the same time), and avoids
    # storage leaks, because files get cleaned up correctly.
    df = Datafile(
        path="my_mandelbrot_file.json",  # File name including extension (and can include subfolders within the dataset)
        local_path_prefix=output_dataset.path,  # TODO set up for the right paths Destination (root of the output dataset folder on the present machine)
        skip_checks=True,  # We haven't created the actual file yet, so it'll definitely fail checks!
        tags=tags,
    )

    # Actually write the contents to the file specified by the Datafile
    # TODO we should write a helper and/or context manager for this, like `df.write_json({})` or `with df('w') as fp:`
    with open(df.full_path, "w") as fp:
        json.dump(
            {"data": data, "layout": layout}, fp, cls=OctueJSONEncoder
        )  # The special encoder just makes it easy to handle numpy arrays

    # And finally we add it to the output
    output_dataset.append(df)



import uuid

from octue.exceptions import InvalidInputException
from octue.utils import gen_uuid


class Identifiable:
    """ Mixin to allow instantiation of a class with a given uuid, or generate one on instantiation

    Prevents setting an id after an object is instantiated.

    Provides a basic str() method which will be overloaded by most Resource classes

    ```
    class MyResource(Identifiable):
        pass

    MyResource().id  # Some generated uuid
    MyResource(id='not_a_uuid')  # Raises exception
    MyResource(id='a10603a0-194c-40d0-a7b7-fcf9952c3690').id  # That same uuid
    ```
    """

    def __init__(self, *args, id=None, **kwargs):
        """ Constructor for Identifiable class
        """
        super().__init__(*args, **kwargs)

        # Store a boolean record of whether this object was created with a previously-existing uuid or was created new.
        self._created = True if id is None else False

        if isinstance(id, uuid.UUID):
            # If it's a uuid, stringify it
            id = str(id)

        elif isinstance(id, str):
            # If it's a string (or something similar which can be converted to UUID) check it's valid
            try:
                id = str(uuid.UUID(id))
            except ValueError:
                raise InvalidInputException(f"Value of id '{id}' is not a valid uuid string or instance of class UUID")

        elif id is not None:
            raise InvalidInputException(
                f"Value of id '{id}' must be a valid uuid string, an instance of class UUID or None"
            )

        self._id = id or gen_uuid()

    def __str__(self):
        return f"{self.__class__.__name__} {self._id}"

    def __repr__(self):
        return self.__str__()

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        raise InvalidInputException(f"You cannot set the id of an already-instantiated {self.__class__.__name__}")



import logging

from octue.cli import AppFrom
from octue.resources.analysis import CLASS_MAP, Analysis
from octue.utils import gen_uuid
from twined import Twine


module_logger = logging.getLogger(__name__)

# Logging format for analysis runs. All handlers should use this logging format, to make logs consistently parseable
LOG_FORMAT = "%(name)s %(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s"


class Runner:
    """ Runs analyses in the app framework

    The Runner class provides a set of configuration parameters for use by your application, together with a range of
    methods for managing input and output file parsing as well as controlling logging.

    :parameter twine: string path to the twine file, or a string containing valid twine json

    :parameter paths: string or dict. If a string, contains a single path to an existing data directory where
    (if not already present), subdirectories 'configuration', 'input', 'tmp', 'log' and 'output' will be created. If a
    dict, it should contain all of those keys, with each of their values being a path to a directory (which will be
    recursively created if it doesn't exist)

    :parameter configuration_values: The strand data. Can be expressed as a string path of a *.json file (relative
    or absolute), as an open file-like object (containing json data), as a string of json data or as an
    already-parsed dict.

    :parameter configuration_manifest: The strand data. Can be expressed as a string path of a *.json file
    (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
    already-parsed dict.

    :parameter skip_file_checks: If true, skip the check that all files in the manifest are present on disc - this
    can be an extremely long process for large datasets.
    """

    def __init__(
        self, twine="twine.json", configuration_values=None, configuration_manifest=None, log_level=logging.INFO
    ):
        """ Constructor for the Runner class
        """

        # Ensure the twine is present and instantiate it
        self.twine = Twine(source=twine)

        # Validate and initialise configuration data
        self.configuration = self.twine.validate(
            configuration_values=configuration_values, configuration_manifest=configuration_manifest, cls=CLASS_MAP,
        )

        # Store the log level (same log level used for all analyses)
        self._log_level = log_level

        # Store analyses. Multiple analysis objects can be created and coexist.
        self.analyses = {}

    def _get_default_handler(self):
        """ Gets a basic console handler set up for logging analyses
        """
        console_handler = logging.StreamHandler()
        console_handler.setLevel(self._log_level)
        formatter = logging.Formatter(LOG_FORMAT)
        console_handler.setFormatter(formatter)

        return console_handler

    def _get_analysis_logger(self, analysis_id, handler=None):
        """ Create a logger specific to the analysis

        :parameter analysis_id: The id of the analysis to get the log for. Should be unique to the analysis
        :type analysis_id: str

        :parameter handler: The handler to use. If None, default console handler will be attached.

        :return: logger named in the pattern `analysis-{analysis_id}`
        :rtype logging.Logger
        """

        handler = handler or self._get_default_handler()
        analysis_logger = logging.getLogger(f"analysis-{analysis_id}")
        analysis_logger.addHandler(handler)

        return analysis_logger

    def run(self, app_src, handler=None, input_values=None, input_manifest=None, credentials=None, children=None):
        """ Run an analysis

        :parameter app_src: Either: an instance of the AppFrom manager class which has a run() method, or
        a function which accepts a single parameter (the instantiated analysis), or a string or path_like pointing
        a string or path_like pointing to an application folder (which should contain an 'app.py' function like the
        templates). This typically points to the run() function defined in the 'app.py' file.
        :type app_src: Union[AppFrom, function, str]

        :parameter input_values: The input_values strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type input_values: Union[str, dict]

        :parameter input_manifest: The input_manifest strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type input_manifest: Union[str, Manifest]

        :parameter credentials: The credentials strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type credentials: Union[str, dict]

        :parameter children: The children strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type children: Union[str, dict]

        :parameter handler: the logging.Handler instance which will be used to handle logs for this analysis run.
        handlers can be created as per the logging cookbook https://docs.python.org/3/howto/logging-cookbook.html but
        should use the format defined above in LOG_FORMAT.
        :type handler: logging.Handler

        :return: None
        """

        inputs = self.twine.validate(
            input_values=input_values,
            input_manifest=input_manifest,
            credentials=credentials,
            children=children,
            cls=CLASS_MAP,
            allow_missing=False,
            allow_extra=False,
        )

        outputs_and_monitors = self.twine.prepare("monitors", "output_values", "output_manifest", cls=CLASS_MAP)

        analysis_id = gen_uuid()
        analysis_logger = self._get_analysis_logger(analysis_id, handler)
        analysis = Analysis(
            id=analysis_id,
            logger=analysis_logger,
            twine=self.twine,
            **self.configuration,
            **inputs,
            **outputs_and_monitors,
        )

        try:
            if hasattr(app_src, "run"):
                app_src.run(analysis)
            elif isinstance(app_src, str):
                with AppFrom(app_src) as app:
                    app.run(analysis)
            else:
                app_src(analysis)

            self.twine.validate(output_values=analysis.output_values)
            self.twine.validate(output_manifest=analysis.output_manifest)

        except Exception as e:
            analysis_logger.error(str(e))
            raise e

        return analysis



import os
import sys

from fractal import fractal
from octue import octue_cli, octue_run, octue_version


@octue_run
def run(analysis):
    """ Your main entrypoint to run the application

    This is the function that gets run each time somebody requests an analysis from the digital twin / data service.
    You should write your own code and call it from here.

    ## The Analysis:

    `analysis` is an instantiated Analysis class object, which you can import here (as shown) or anywhere else in your
    code. It contains:
        - ``configuration_values``, which have been validated against the twine
        - ``configuration_manifest``, a Manifest instance whose contents have been validated and whose files have been
           checked (files are checked to be present and, if a `sha` field is given in the manifest, their
           contents checked to match the sha)
        - ``input_values``, which have been validated against the twine
        - ``input_manifest``, a Manifest instance whose contents have been validated and whose files have been
           checked (files are checked to be present and, if a `sha` field is given in the manifest, their
           contents checked to match the sha)
        - ``output_values``, dict which can be added to as required (on completion, it will be validated
           against the twine and returned to the requester)
        - ``output_manifest``, a Manifest instance to which newly created datasets and files should be added (on
           completion, the presence of the files will be checked, their shas calculated and they'll be returned to the
           requester or uploaded)
        - ``children``, a dict of Child objects allowing you to access child twins/services
        - ``credentials``, a dict of Credential objects

    """

    # You can use the attached logger to record debug statements, general information, warnings or errors
    # analysis.logger.info(f"The input directory is {analysis.input_dir}")
    # analysis.logger.info(f"The output directory is {analysis.output_dir}")
    # analysis.logger.info(f"The tmp directory, where you can store temporary files or caches, is {analysis.tmp_dir}")

    # Print statements will get logged (stdout and stderr are mirrored to the log files so you don't miss anything)...
    print("Hello! The app is running!")  # noqa: T001

    # ... but we encourage you to use the attached logger, which handles sending logs to remote services and allows them
    # to be viewed with twined server
    analysis.logger.info(
        "The logger can be used for capturing different 'levels' of statement - for debug, info, warnings or errors."
    )

    # You can access any of the configuration or input values, anywhere in your code, from the analysis object
    analysis.logger.info(f"The maximum number of iterations will be {analysis.configuration_values}")

    # Run the code
    fractal(analysis)


@octue_version
def version():
    """ Returns the version number of the application
    """

    # Top Tip:
    # For all Octue internal apps, we simply return the git revision of the code.
    # Every single commit creates a new version, we can always check out the exact version of the code that ran, and we
    # can quickly look up the version state and history on github when we have to debug an app. Sweet!
    version_no = os.system("git rev-parse HEAD")

    # Return the version number as a string
    return version_no


# If running from an IDE or test console, it'll run this file rather than calling the application from the CLI...
# In that case we pass arguments through the CLI just as if it were called from the command line.
if __name__ == "__main__":

    # Invoke the CLI to process the arguments, set up an analysis and run it
    args = sys.argv[1:] if len(sys.argv) > 1 else []
    octue_cli(args)



import os
from setuptools import setup


def git_version():
    return os.system("git rev-parse HEAD")


setup(
    name="template-python-fractal",  # Change this if you want to make your module installable as a library
    version=git_version(),
    py_modules=["app"],
    entry_points="""
    [console_scripts]
    octue-app=app:octue_app
    """,
)



import os

from octue import exceptions
from .isfolder import isfolder


FOLDERS = (
    "configuration",
    "input",
    "log",
    "tmp",
    "output",
)


def from_path(path_hints, folders=FOLDERS):
    """ NOT IMPLEMENTED YET - Helper to find paths to individual configurations from hints
    TODO Fix this
    """
    # Set paths
    paths = dict()
    if isinstance(path_hints, str):
        if not os.path.isdir(path_hints):
            raise exceptions.FolderNotFoundException(f"Specified data folder '{path_hints}' not present")

        paths = dict([(folder, os.path.join(path_hints, folder)) for folder in folders])

    else:
        if (
            not isinstance(paths, dict)
            or (len(paths.keys()) != len(folders))
            or not all([k in folders for k in paths.keys()])
        ):
            raise exceptions.InvalidInputException(
                f"Input 'paths' should be a dict containing directory paths with the following keys: {folders}"
            )

    # Ensure paths exist on disc??
    for folder in FOLDERS:
        isfolder(paths[folder], make_if_absent=True)



class OctueSDKException(Exception):
    """ All exceptions from this library inherit from here
    """


class InvalidInputException(OctueSDKException, ValueError):
    """Raise when an object is instantiated or a function called with invalid inputs
    """


class FileNotFoundException(InvalidInputException, FileNotFoundError):
    """ Raise when a required folder (e.g. <data_dir>/input) cannot be found
    """


class ProtectedAttributeException(OctueSDKException, KeyError):
    """ Raise when a user attempts to set an attribute whose value should be protected
    """


class FolderNotFoundException(InvalidInputException):
    """ Raise when a multi manifest can not be refined to a single manifest in a search
    """


class ManifestNotFoundException(InvalidInputException):
    """ Raise when a multi manifest can not be refined to a single manifest in a search
    """


class InvalidManifestException(InvalidInputException):
    """Raise when a manifest loaded from JSON does not pass validation
    """


class InvalidManifestTypeException(InvalidManifestException):
    """Raised when user attempts to create a manifest of a type other than 'input', 'output' or 'build'
    """


class InvalidOctueFileTypeException(OctueSDKException):
    """Raised when you attempt to register a file type in the results manifest that Octue doesn't know about
    """


class InvalidFilePointerException(OctueSDKException):
    """Raised when you attempt to create a Datafile with an object that is not file-like.
    """


class NotImplementedYetException(OctueSDKException):
    """Raised when you attempt to use a function whose high-level API is in place, but which is not implemented yet
    """


class UnexpectedNumberOfResultsException(OctueSDKException):
    """ Raise when searching for a single data file (or a particular number of data files) and the number of results
    exceeds that expected
    """


class BrokenSequenceException(OctueSDKException):
    """ Raise when filtering for a sequence of files whose numbering does not start at 0, or does not monotonically
    increase
    """


class InvalidTagException(OctueSDKException, ValueError):
    """ Raise when a tag applied to a data file or dataset
    """



from json import JSONDecoder

from octue.resources import Datafile, Dataset, Manifest


def default_object_hook(obj):
    """ A hook to convert default json objects into their Datafile, Dataset or Manifest class as appropriate
    """

    # object hooks are called whenever a json object is created. When nested, this is done from innermost (deepest
    # nesting) out so it's safe to work at multiple levels here

    if "files" in obj:
        files = [Datafile(**df) for df in obj.pop("files")]
        return {**obj, "files": files}

    if "datasets" in obj:
        datasets = [Dataset(**ds) for ds in obj.pop("datasets")]
        return Manifest(**obj, datasets=datasets)

    return obj


class OctueJSONDecoder(JSONDecoder):
    """ A JSON Decoder to convert default json objects into their Datafile, Dataset or Manifest classes as appropriate
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.object_hook = self.object_hook or default_object_hook



import re

from octue.exceptions import InvalidTagException


# A tag starts and ends with a character in [a-z] or [0-9]. It can contain the colon discriminator or hyphens.
# Empty strings are also valid.
# Valid examples:
#    system:32
#    angry-marmaduke
#    mega-man:torso:component:12
TAG_PATTERN = re.compile(r"^$|^[a-z0-9][a-z0-9:\-]*(?<![:-])$")


class TagGroup:
    """ Class to handle a group of tags as a string.
    """

    def __init__(self, tags):
        """ Construct a TagGroup
        """
        if tags is None:
            tags = []

        elif isinstance(tags, str):
            tags = tags.strip().split()

        elif isinstance(tags, TagGroup):
            # Use the serialise method as a deepcopy
            tags = tags.serialise().split()

        # Tags can be some other iterable than a list, but each tag must be a string
        elif not hasattr(tags, "__iter__"):
            raise InvalidTagException(
                "Tags must be expressed as a whitespace-delimited string or an iterable of strings"
            )

        self._tags = self._clean(tags)

    @staticmethod
    def _clean(tags):
        """ Private method to clean up an iterable of tags into a list of cleaned tags
        """
        # Check they're strings
        if not all(isinstance(tag, str) for tag in tags):
            raise InvalidTagException(
                "Tags must be expressed as a whitespace-delimited string or an iterable of strings"
            )

        # Strip leading and trailing whitespace, ensure they match the regex
        cleaned_tags = []
        for tag in tags:
            cleaned_tag = tag.strip()
            if not re.match(TAG_PATTERN, cleaned_tag):
                raise InvalidTagException(
                    "Tags must contain only characters 'a-z', '0-9', ':' and '-'. They must not start with '-' or ':'."
                )
            cleaned_tags.append(cleaned_tag)

        return cleaned_tags

    def __str__(self):
        """ Represents tags as a space delimited string
        """
        return self.serialise()

    def serialise(self):
        """ Serialises tags as a space delimited string, NOT as a list. Strips end whitespace.
        """
        return " ".join(self._tags).strip()

    def startswith(self, value):
        """ Implement a startswith method that returns true if any of the tags starts with value
        """
        return any(tag.startswith(value) for tag in self._tags)

    def endswith(self, value):
        """ Implement an endswith method that returns true if any of the tags endswith value
        """
        return any(tag.endswith(value) for tag in self._tags)

    def contains(self, value):
        """ Implement a contains method that returns true if any of the tags contains value
        """
        return any(value in tag for tag in self._tags)

    def __contains__(self, value):
        """ Returns true if any of the tags exactly matches value, allowing test like `if 'a' in TagGroup('a b')`
        """
        return any(value == tag for tag in self._tags)

    def add_tags(self, *args):
        """ Adds one or more new tag strings to the object tags. New tags will be cleaned and validated.
        """
        self._tags += self._clean(args)


class Taggable:
    """ A mixin class allowing objects to be tagged
    """

    def __init__(self, tags=None, **kwargs):
        """ Constructor for Taggable mixins
        """
        super().__init__(**kwargs)
        self._tags = TagGroup(tags)

    def add_tags(self, *args):
        """ Adds one or more new tag strings to the object tags. New tags will be cleaned and validated.
        """
        self._tags.add_tags(*args)

    @property
    def tags(self):
        """ Space delimited tag string
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """ Overwrite any existing tag group and assign new tags
        """
        self._tags = TagGroup(tags)



import logging

from octue.exceptions import ProtectedAttributeException
from octue.mixins import Identifiable, Loggable, Serialisable, Taggable
from octue.resources.manifest import Manifest
from twined import ALL_STRANDS, Twine


module_logger = logging.getLogger(__name__)


# Map strand names to class which we expect Twined to instantiate for us
CLASS_MAP = {"configuration_manifest": Manifest, "input_manifest": Manifest, "output_manifest": Manifest}


class Analysis(Identifiable, Loggable, Serialisable, Taggable):
    """ Analysis class, holding references to all input and output data

    ## The Analysis Instance

    An Analysis instance is unique to a specific computation analysis task, however large or small, run at a specific
    time. It will be created by the task runner (which will have validated incoming data already - Analysis() doesn't
    do any validation).

    It holds references to all config, input and output data, logs, connections to child twins, credentials, etc, so
    should be referred to from your code to get those items.

    It's basically the "Internal API" for your data service - a single point of contact where you can get or update
    anything you need.

    Analyses are instantiated at the top level of your app/service/twin code and you can import the instantiated
    object from there (see the templates for examples)

    :parameter twine: Twine instance or json source
    :parameter configuration_values: see Runner.run() for definition
    :parameter configuration_manifest: see Runner.run() for definition
    :parameter input_values: see Runner.run() for definition
    :parameter input_manifest: see Runner.run() for definition
    :parameter credentials: see Runner.run() for definition
    :parameter monitors: see Runner.run() for definition
    :parameter output_values: see Runner.run() for definition
    :parameter output_manifest: see Runner.run() for definition
    :parameter id: Optional UUID for the analysis
    :parameter logger: Optional logging.Logger instance attached to the analysis
    """

    def __init__(self, twine, **kwargs):
        """ Constructor of Analysis instance
        """

        # Instantiate the twine (if not already) and attach it to self
        if not isinstance(twine, Twine):
            twine = Twine(source=twine)

        self.twine = twine

        # Pop any possible strand data sources before init superclasses (and tie them to protected attributes)
        strand_kwargs = dict((name, kwargs.pop(name, None)) for name in ALL_STRANDS)
        for strand_name, strand_data in strand_kwargs.items():
            self.__setattr__(f"_{strand_name}", strand_data)

        # Init superclasses
        super().__init__(**kwargs)

    def __setattr__(self, name, value):
        """ Override setters for protected attributes (the strand contents may change, but the strands themselves
        shouldn't be changed after instantiation)
        """
        if name in ALL_STRANDS:
            raise ProtectedAttributeException(f"You cannot set {name} on an instantiated Analysis")

        super().__setattr__(name, value)

    def __getattr__(self, name):
        """ Override public getters to point to protected attributes (the strand contents may change, but the strands
        themselves shouldn't be changed after instantiation)
        """
        if name in ALL_STRANDS:
            return getattr(self, f"_{name}")



import logging

from octue.exceptions import InvalidInputException
from octue.mixins import Identifiable, Loggable, Serialisable
from .dataset import Dataset


module_logger = logging.getLogger(__name__)


class Manifest(Serialisable, Loggable, Identifiable):
    """ A representation of a manifest, which can contain multiple datasets
    This is used to manage all files coming into (or leaving), a data service for an analysis at the
    configuration, input or output stage.
    """

    def __init__(self, id=None, logger=None, **kwargs):
        """ Construct a Manifest
        """
        super().__init__(id=id, logger=logger)
        self.datasets = kwargs.pop("datasets", list())
        self.keys = kwargs.pop("keys", dict())
        self.__dict__.update(**kwargs)

        # TODO we need to add keys to the manifest file schema so that we know what dataset(s) map to what keys

    def get_dataset(self, key):
        """ Gets a dataset by its key name (as defined in the twine)
        """
        idx = self.keys.get(key, None)
        if idx is None:
            raise InvalidInputException(
                f"Attempted to fetch unknown dataset '{key}' from Manifest. Allowable keys are: {list(self.keys.keys())}"
            )

        return self.datasets[idx]

    def prepare(self, data):
        """ Prepare new manifest from a manifest_spec
        """
        if len(self.datasets) > 0:
            raise InvalidInputException("You cannot `prepare()` a manifest already instantiated with datasets")

        for idx, dataset_spec in enumerate(data):

            self.keys[dataset_spec["key"]] = idx
            # TODO generate a unique name based on the filter key, tag datasets so that the tag filters in the spec
            #  apply automatically and generate a description of the dataset
            self.datasets.append(Dataset(logger=self.logger))

        return self



import os
from shutil import copytree
from pkg_resources import resource_filename, resource_listdir

from octue import exceptions


# TODO add ONLINE_TEMPLATES and combine to AVAILABLE_TEMPLATES, to enable us to seamlessly deliver templates that are
#  more complex (or have lots of data that we don't want packaged with the module).

PACKAGED_TEMPLATES = tuple(filter(lambda name: name.startswith("template-"), resource_listdir("octue", "templates")))


def copy_template(template_name, destination_dir="."):
    """ Copies one of the application templates from the octue/templates to a destination directory (current dir by default)

    :parameter template_name: The name of the template to copy, must be one of AVAILABLE_TEMPLATES.
    :type template_name: str

    :parameter destination_dir: The destination directory to copy to, default "." (current directory)
    :type destination_dir: path-like
    """

    if template_name not in PACKAGED_TEMPLATES:
        raise exceptions.InvalidInputException(
            f"Unknown template name '{template_name}', try one of {PACKAGED_TEMPLATES}"
        )

    resource = resource_filename("octue.templates", template_name)
    dest = os.path.join(destination_dir, template_name)
    copytree(resource, dest)



import hashlib
import logging
import os
import time

from octue.exceptions import FileNotFoundException, InvalidInputException
from octue.mixins import Identifiable, Loggable, Serialisable, Taggable
from octue.utils import isfile


module_logger = logging.getLogger(__name__)


class Datafile(Taggable, Serialisable, Loggable, Identifiable):
    """ Class for representing data files on the Octue system

    Files in a manifest look like this:

        {
          "path": "input/datasets/7ead7669/file_1.csv",
          "cluster": 0,
          "sequence": 0,
          "extension": "csv",
          "tags": "",
          "posix_timestamp": 0,
          "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86",
          "last_modified": "2019-02-28T22:40:30.533005Z",
          "size_bytes": 59684813,
          "sha-512/256": "somesha"
        },

    :parameter local_path_prefix: A path, specific to the present local system, to the directory containing the dataset
    in which this file resides. Default is the current working directory.
    :type local_path_prefix: str

    :parameter path: The path of this file, relative to the local_path_prefix (which may have a folder structure within it)
    :type path: str

    :parameter logger: A logger instance to which operations with this datafile will be logged. Defaults to the module logger.
    :type logger: logging.Logger

    :parameter id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
    :type id: str

    :parameter cluster: The cluster of files, within a dataset, to which this belongs (default 0)
    :type cluster: int

    :parameter sequence: A sequence number of this file within its cluster (if sequences are appropriate)
    :type sequence: int

    :parameter tags: Space-separated string of tags relevant to this file
    :type tags: str

    :parameter posix_timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when it
    was created but could relate to a relevant time point for the data
    :type posix_timestamp: number
    """

    def __init__(
        self,
        id=None,
        logger=None,
        local_path_prefix=".",
        path=None,
        cluster=0,
        sequence=None,
        tags=None,
        posix_timestamp=None,
        skip_checks=True,
        **kwargs,
    ):
        """ Construct a datafile
        """
        super().__init__(id=id, logger=logger, tags=tags)

        self.cluster = cluster

        self.sequence = sequence
        self.posix_timestamp = posix_timestamp or time.time()

        if path is None:
            raise InvalidInputException("You must supply a valid 'path' argument")
        else:
            path = str(path).lstrip("\\/")

        # Strip to ensure path is always expressed as relative
        self.path = path

        # Replace current directory specifier in the prefix with an absolute path
        self.local_path_prefix = str(os.path.abspath(local_path_prefix))

        # Set up the file extension or get it from the file path if none passed
        self.extension = self._get_extension_from_path()

        # Run integrity checks on the file
        if not skip_checks:
            self.check(**kwargs)

    def _get_extension_from_path(self, path=None):
        """ Gets extension of a file, either from a provided file path or from self.path field
        """
        path = path or self.path
        return os.path.splitext(path)[-1].strip(".")

    def _get_sha_256(self):
        """ Calculate the SHA256 hash string of the file
        """
        sha256_hash = hashlib.sha256()
        with open(self.full_path, "rb") as f:
            # Read and update hash string value in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)

        return sha256_hash.hexdigest()

    @property
    def name(self):
        return str(os.path.split(self.path)[-1])

    @property
    def full_path(self):
        return os.path.join(self.local_path_prefix, self.path)

    @property
    def last_modified(self):
        return os.path.getmtime(self.full_path)

    @property
    def size_bytes(self):
        return os.path.getsize(self.full_path)

    @property
    def sha_256(self):
        return self._get_sha_256()

    def check(self, size_bytes=None, sha=None, last_modified=None, extension=None):
        """ Check file presence and integrity
        """
        # TODO Check consistency of size_bytes input against self.size_bytes property for a file if we have one
        # TODO Check consistency of sha against file contents if we have a file
        # TODO Check consistency of last_modified date

        if (extension is not None) and not self.path.endswith(extension):
            raise InvalidInputException(
                f"Extension provided ({extension}) does not match file extension (from {self.path}). Pass extension=None to set extension from filename automatically."
            )

        if not self.exists():
            raise FileNotFoundException(f"No file found at {self.full_path}")

    def exists(self):
        """ Returns true if the datafile exists on the current system, false otherwise
        """
        return isfile(self.full_path)
